local require = require
require("eclass")

local nsenter = require("nsenter")
local elfmap = require("elfmap")
local cjson = require("cjson.safe")
local fastKsym = require("haWork.fastKsym")
local CbaseFork = require("haWork.baseFork")
local demangle_ffi = require("haWork.demangleFfi")
local dm_type, dm_api = demangle_ffi.type, demangle_ffi.api

local pcall = pcall
local print = print
local class = class
local tostring = tostring
local jencode = cjson.encode
local jdecode = cjson.decode
local match = string.match
local format = string.format
local c_new = dm_type.new
local c_string = dm_type.string
local demangle = dm_api.demangle
local loadKsyms = fastKsym.loadSyms
local collectgarbage = collectgarbage

local ChaElfSyms = class("ChaElfSyms", CbaseFork)
cjson.encode_escape_forward_slash(false)

function ChaElfSyms:_init_()
    CbaseFork._init_(self)
end

local function extractBeforeParentheses(s)
    local result = match(s, "^(.-)%(")
    return result or s
end

local function removeNestedAngleBrackets(s)
    local prev_str
    repeat
        prev_str = s
        s = s:gsub("%b<>", "")
    until s == prev_str
    local result = match(s, "^(.-)<")
    return result or s
end

local function removeSpacesUnprintable(s)
    s = s:gsub("%s+", "")
    s = s:gsub("[%z\1-\31\127]+", "")
    return s
end

local function newMap(elf)
    local map = elfmap.new(elf)
    local cnt = 1
    local symbols = {}
    local stopEnd
    local buffSize = 256
    local buff = c_new("char[?]", buffSize)
    local bias = map:bias()
    local s

    for _, start, stop, symbol in map:maps() do
        if demangle(symbol, buff, buffSize) == 0 then  --c++ may contain c symbol. can not use call back function
            s = c_string(buff, buffSize)
            s = extractBeforeParentheses(s)
            s = removeNestedAngleBrackets(s)
            s = removeSpacesUnprintable(s)
        else
            s = symbol
        end
        symbols[cnt] = {start - bias, s}
        stopEnd = stop
        cnt = cnt + 1
    end
    if cnt <= 1 then
        return "{}"
    end
    symbols[cnt] = {stopEnd - bias + 1, "-"}
    -- print(format("newMap: %s, start: %d, stop: %d, bias: %d, count:%d", elf, tonumber(symbols[1][1]), tonumber(symbols[cnt][1]), bias, #symbols))
    return jencode(symbols)
end

function ChaElfSyms:_main(r, w)
    local ns = nsenter.new()

    while true do
        local s = self:pipeRead(r)
        local conf = jdecode(s)
        s = ""
        if conf and conf.pid and conf.elf then
            local res, err = pcall(ns.enter, ns, conf.pid, "mnt")
            if res then
                local ok, jRes = pcall(newMap, conf.elf)
                if ok then
                    s = jRes
                else
                    print("newMap failed", conf.pid, conf.elf, err)
                    s = "{}"
                end
                ns:exit()
            else
                print("nsenter failed", err, conf.pid)
            end
        else
            print("elfdump exit.")
            break
        end
        if #s == 0 then
            print("bad elf for", conf.pid, conf.elf)
        end
        self:pipeWrite(w, s)
        collectgarbage("collect")
    end
    print("fork exit.")
end

function ChaElfSyms:getElfs(pid, elf)
    if elf == "/proc/kallsyms" then
        local arr, bias = loadKsyms()
        return {arr = arr, bias = tostring(bias)}
    end
    local send = {
        elf = elf,
        pid = pid,
    }
    self:pwrite(jencode(send))
    local s = self:pread()
    return jdecode(s)
end

return ChaElfSyms
